# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import os
import pickle
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
import pytest
from tests.causal_manager_validator import validate_causal
from tests.common_utils import create_iris_data
from tests.counterfactual_manager_validator import validate_counterfactual
from tests.error_analysis_validator import (setup_error_analysis,
                                            validate_error_analysis)
from tests.explainer_manager_validator import (setup_explainer,
                                               validate_explainer)

from rai_test_utils.datasets.tabular import (
    create_binary_classification_dataset, create_cancer_data,
    create_housing_data)
from rai_test_utils.models.model_utils import (create_models_classification,
                                               create_models_regression)
from rai_test_utils.models.sklearn import \
    create_complex_classification_pipeline
from rai_test_utils.utilities import is_valid_uuid
from raiutils.models import ModelTask
from responsibleai import RAIInsights
from responsibleai._internal.constants import (ManagerNames,
                                               SerializationAttributes)
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.feature_metadata import FeatureMetadata

LABELS = 'labels'


class ManagerParams:
    # Counterfactual
    DESIRED_CLASS = 'desired_class'
    DESIRED_RANGE = 'desired_range'
    FEATURE_IMPORTANCE = 'feature_importance'

    # Causal
    TREATMENT_FEATURES = 'treatment_features'
    MAX_CAT_EXPANSION = 'max_cat_expansion'


class TestRAIInsights(object):

    @pytest.mark.parametrize('manager_type', [ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.CAUSAL,
                                              ManagerNames.EXPLAINER])
    def test_rai_insights_iris(self, manager_type):
        X_train, X_test, y_train, y_test, feature_names, classes = \
            create_iris_data()
        X_train_after_drop = X_train.drop(columns=['petal length'])
        models = create_models_classification(X_train_after_drop, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: [feature_names[0]],
            ManagerParams.DESIRED_CLASS: 0,
            ManagerParams.FEATURE_IMPORTANCE: True
        }

        feature_metadata = FeatureMetadata(
            identity_feature_name='sepal length',
            dropped_features=['petal length'])
        for model in models:
            run_rai_insights(model, X_train, X_test, LABELS, None,
                             manager_type, manager_args, classes,
                             feature_metadata=feature_metadata)

    @pytest.mark.parametrize('manager_type', [ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.EXPLAINER])
    def test_rai_insights_cancer(self, manager_type):
        X_train, X_test, y_train, y_test, _, classes = \
            create_cancer_data(return_dataframe=True)
        models = create_models_classification(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        manager_args = {
            ManagerParams.DESIRED_CLASS: 'opposite',
            ManagerParams.FEATURE_IMPORTANCE: False
        }

        for model in models:
            run_rai_insights(model, X_train, X_test, LABELS, None,
                             manager_type, manager_args, classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER])
    def test_rai_insights_binary(self, manager_type):
        X_train, y_train, X_test, y_test, classes = \
            create_binary_classification_dataset()

        models = create_models_classification(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: ['col0']
        }

        for model in models:
            run_rai_insights(model, X_train, X_test, LABELS, None,
                             manager_type, manager_args,
                             classes=classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL])
    @pytest.mark.parametrize('add_boolean', [True, False])
    def test_rai_insights_binary_mixed_types(
            self, manager_type, add_boolean, adult_data):

        data_train, data_test, y_train, y_test, categorical_features, \
            continuous_features, target_name, classes, \
            feature_columns, feature_range_keys = \
            adult_data

        if add_boolean:
            data_train = data_train.copy()
            data_test = data_test.copy()
            data_train['is_adult'] = data_train['age'] >= 18
            data_test['is_adult'] = data_test['age'] >= 18
            categorical_features = categorical_features + ['is_adult']

        X_train = data_train.drop([target_name], axis=1)
        dropped_feature = 'education'
        X_train_after_drop = X_train.drop([dropped_feature], axis=1)
        categorical_features_after_drop = categorical_features.copy()
        categorical_features_after_drop.remove(dropped_feature)

        model = create_complex_classification_pipeline(
            X_train_after_drop, y_train, continuous_features,
            categorical_features_after_drop)
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: ['age', 'hours_per_week'],
            ManagerParams.DESIRED_CLASS: 'opposite',
            ManagerParams.FEATURE_IMPORTANCE: False
        }

        feature_metadata = FeatureMetadata(
            dropped_features=[dropped_feature])
        run_rai_insights(model, data_train, data_test, target_name,
                         categorical_features,
                         manager_type, manager_args,
                         classes=classes, feature_metadata=feature_metadata)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL])
    def test_rai_insights_no_model(self, manager_type):

        X_train, y_train, X_test, y_test, classes = \
            create_binary_classification_dataset()

        X_train[LABELS] = y_train
        X_test[LABELS] = y_test
        model = None
        manager_args = {
            ManagerParams.TREATMENT_FEATURES: ['col0']
        }

        run_rai_insights(model, X_train, X_test, LABELS, None,
                         manager_type, manager_args,
                         classes=classes)

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.COUNTERFACTUAL,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.ERROR_ANALYSIS])
    def test_rai_insights_housing(self, manager_type):
        X_train, X_test, y_train, y_test, feature_names = \
            create_housing_data()
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)
        models = create_models_regression(X_train, y_train)
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test

        manager_args = {
            ManagerParams.DESIRED_RANGE: [3, 5],
            ManagerParams.TREATMENT_FEATURES: ['AveRooms'],
            ManagerParams.MAX_CAT_EXPANSION: 12,
            ManagerParams.FEATURE_IMPORTANCE: True
        }

        for model in models:
            run_rai_insights(model, X_train, X_test, LABELS, [],
                             manager_type, manager_args)


def run_rai_insights(model, train_data, test_data, target_column,
                     categorical_features, manager_type,
                     manager_args=None, classes=None,
                     feature_metadata=None):
    if manager_args is None:
        manager_args = {}

    if classes is not None:
        task_type = ModelTask.CLASSIFICATION
    else:
        task_type = ModelTask.REGRESSION

    if manager_type == ManagerNames.COUNTERFACTUAL:
        feature_importance = manager_args.get(ManagerParams.FEATURE_IMPORTANCE)
        if feature_importance:
            test_data = test_data[0:20]
        else:
            test_data = test_data[0:1]

    if model is None:
        with pytest.warns(
                UserWarning,
                match='INVALID-MODEL-WARNING: '
                      'No valid model is supplied. '
                      'The explanations, error analysis and '
                      'counterfactuals may not work'):
            rai_insights = RAIInsights(
                model, train_data, test_data,
                target_column,
                categorical_features=categorical_features,
                task_type=task_type,
                feature_metadata=feature_metadata)
    else:
        rai_insights = RAIInsights(
            model, train_data, test_data,
            target_column,
            categorical_features=categorical_features,
            task_type=task_type,
            feature_metadata=feature_metadata)

    if manager_type == ManagerNames.EXPLAINER:
        setup_explainer(rai_insights)
    elif manager_type == ManagerNames.ERROR_ANALYSIS:
        setup_error_analysis(rai_insights)

    validate_rai_insights(rai_insights, train_data, test_data, target_column,
                          task_type, categorical_features, feature_metadata)

    if manager_type == ManagerNames.CAUSAL:
        treatment_features = manager_args.get(ManagerParams.TREATMENT_FEATURES)
        max_cat_expansion = manager_args.get(ManagerParams.MAX_CAT_EXPANSION)
        validate_causal(rai_insights, train_data, target_column,
                        treatment_features, max_cat_expansion)
    elif manager_type == ManagerNames.COUNTERFACTUAL:
        desired_class = manager_args.get(ManagerParams.DESIRED_CLASS)
        desired_range = manager_args.get(ManagerParams.DESIRED_RANGE)
        feature_importance = manager_args.get(ManagerParams.FEATURE_IMPORTANCE)
        validate_counterfactual(rai_insights,
                                desired_class=desired_class,
                                desired_range=desired_range,
                                feature_importance=feature_importance)
    elif manager_type == ManagerNames.ERROR_ANALYSIS:
        validate_error_analysis(rai_insights)
    elif manager_type == ManagerNames.EXPLAINER:
        validate_explainer(rai_insights, train_data, test_data, classes)

    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir) / 'rai_test_path'
        # save the rai_insights
        rai_insights.save(path)

        # Validate the common set of state produced when rai insights
        # are saved on the disk.
        validate_common_state_directories(path, task_type)

        # Validate the directory structure of the state saved
        # by the managers.
        validate_component_state_directory(path, manager_type)

        # load the rai_insights
        rai_insights = RAIInsights.load(path)

        validate_rai_insights(
            rai_insights, train_data, test_data,
            target_column, task_type, categorical_features,
            feature_metadata)

        if manager_type == ManagerNames.ERROR_ANALYSIS:
            validate_error_analysis(rai_insights)
            # validate adding new reports after deserialization works
            setup_error_analysis(rai_insights, max_depth=4)
            validate_error_analysis(rai_insights, expected_reports=2)
        elif manager_type == ManagerNames.EXPLAINER:
            validate_explainer(rai_insights, train_data, test_data, classes)
            # validate adding new explainer config after deserialization works
            setup_explainer(rai_insights)
            validate_explainer(rai_insights, train_data, test_data, classes)


def validate_common_state_directories(path, task_type):
    all_other_files = os.listdir(path)
    assert SerializationAttributes.RAI_VERSION_JSON in all_other_files
    assert SerializationAttributes.META_JSON in all_other_files
    assert SerializationAttributes.MODEL_PKL in all_other_files

    model = None
    with open(path / SerializationAttributes.MODEL_PKL, 'rb') as file:
        model = pickle.load(file)

    predictions_path = path / SerializationAttributes.PREDICTIONS_DIRECTORY
    assert predictions_path.exists()
    all_predictions_files = os.listdir(predictions_path)
    if model is not None:
        assert SerializationAttributes.PREDICT_JSON in all_predictions_files
        if task_type == ModelTask.CLASSIFICATION:
            assert SerializationAttributes.PREDICT_PROBA_JSON in \
                all_predictions_files
        else:
            assert SerializationAttributes.PREDICT_PROBA_JSON not in \
                all_predictions_files
    else:
        assert len(all_predictions_files) == 0

    data_path = path / SerializationAttributes.DATA_DIRECTORY
    assert data_path.exists()
    all_data_files = os.listdir(data_path)
    assert "train.json" in all_data_files
    assert "traindtypes.json" in all_data_files
    assert "test.json" in all_data_files
    assert "testdtypes.json" in all_data_files


def validate_component_state_directory(path, manager_type):
    all_dirs = os.listdir(path)
    assert manager_type in all_dirs
    all_component_paths = os.listdir(path / manager_type)
    for component_path in all_component_paths:
        # Test if the component directory has UUID structure
        is_valid_uuid(component_path)
        dm = DirectoryManager(path / manager_type, component_path)

        config_path = dm.get_config_directory()
        data_path = dm.get_data_directory()
        generators_path = dm.get_generators_directory()

        if manager_type == ManagerNames.EXPLAINER:
            assert not config_path.exists()
            assert data_path.exists()
            assert not generators_path.exists()
        elif manager_type == ManagerNames.COUNTERFACTUAL:
            assert config_path.exists()
            assert data_path.exists()
            assert generators_path.exists()
        elif manager_type == ManagerNames.ERROR_ANALYSIS:
            assert config_path.exists()
            assert data_path.exists()
            assert not generators_path.exists()
        elif manager_type == ManagerNames.CAUSAL:
            assert not config_path.exists()
            assert data_path.exists()
            assert not generators_path.exists()


def validate_rai_insights(
    rai_insights,
    train_data,
    test_data,
    target_column,
    task_type,
    categorical_features,
    feature_metadata=None
):

    pd.testing.assert_frame_equal(rai_insights.train, train_data)
    pd.testing.assert_frame_equal(rai_insights.test, test_data)
    assert rai_insights.target_column == target_column
    assert rai_insights.task_type == task_type
    assert rai_insights.categorical_features == (categorical_features or [])
    expected_length = 0
    if categorical_features is not None:
        expected_length = len(categorical_features)
    assert len(rai_insights.categorical_features) == expected_length
    assert len(rai_insights._categories) == expected_length
    assert len(rai_insights._categorical_indexes) == expected_length
    assert len(rai_insights._category_dictionary) == expected_length
    for ind_data in rai_insights._string_ind_data:
        assert len(ind_data) == expected_length

    if feature_metadata is not None:
        if feature_metadata.dropped_features is not None and \
                categorical_features is not None:
            assert len(rai_insights.get_categorical_features_after_drop()) == \
                len(categorical_features) - \
                len(feature_metadata.dropped_features)
    else:
        if categorical_features is not None:
            assert len(rai_insights.get_categorical_features_after_drop()) == \
                len(categorical_features)
        else:
            assert len(rai_insights.get_categorical_features_after_drop()) == 0

    if rai_insights.model is None:
        assert rai_insights._predict_output is None
        assert rai_insights._predict_proba_output is None
    else:
        assert rai_insights._predict_output is not None
        if task_type == ModelTask.CLASSIFICATION:
            assert rai_insights._predict_proba_output is not None
            assert isinstance(rai_insights._predict_proba_output, np.ndarray)
            assert len(rai_insights._predict_proba_output.tolist()[0]) == \
                len(rai_insights._classes)

    if task_type == ModelTask.CLASSIFICATION:
        classes = train_data[target_column].unique()
        classes.sort()
        np.testing.assert_array_equal(rai_insights._classes,
                                      classes)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from enum import Enum

import numpy as np
import pandas as pd
import pytest
from ml_wrappers.model.predictions_wrapper import \
    PredictionsModelWrapperClassification
from tests.common_utils import create_iris_data

from rai_test_utils.models.sklearn import (
    create_complex_classification_pipeline, create_sklearn_svm_classifier)
from raiutils.exceptions import UserConfigValidationException
from responsibleai import RAIInsights
from responsibleai._internal.constants import ManagerNames
from responsibleai.feature_metadata import FeatureMetadata

LABELS = 'labels'


class MISSING_VALUE(Enum):
    NO_MISSING_VALUES = 1
    TRAIN_ONLY_MISSING_VALUES = 2
    TEST_ONLY_MISSING_VALUES = 3
    BOTH_TRAIN_TEST_MISSING_VALUES = 4


class TestRAIInsightsMissingValues(object):

    def test_model_does_not_handle_missing_values(self):
        X_train, X_test, y_train, y_test, feature_names, classes = \
            create_iris_data()

        model = create_sklearn_svm_classifier(X_train, y_train)
        X_train.at[1, 'sepal length'] = np.nan
        X_test.at[1, 'sepal length'] = np.nan
        X_train[LABELS] = y_train
        X_test[LABELS] = y_test

        with pytest.raises(
                UserConfigValidationException,
                match='The passed model cannot be '
                      'used for getting predictions via predict'):
            RAIInsights(model, X_train, X_test,
                        LABELS, task_type="classification")

    @pytest.mark.parametrize('manager_type', [ManagerNames.CAUSAL,
                                              ManagerNames.ERROR_ANALYSIS,
                                              ManagerNames.EXPLAINER,
                                              ManagerNames.COUNTERFACTUAL])
    @pytest.mark.parametrize('categorical_missing_values', [True, False])
    @pytest.mark.parametrize('missing_value_combination', [
        MISSING_VALUE.NO_MISSING_VALUES,
        MISSING_VALUE.TRAIN_ONLY_MISSING_VALUES,
        MISSING_VALUE.TEST_ONLY_MISSING_VALUES,
        MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES
    ])
    @pytest.mark.parametrize('wrapper', [True, False])
    def test_model_handles_missing_values(
            self, manager_type, adult_data,
            categorical_missing_values,
            missing_value_combination,
            wrapper):

        data_train, data_test, y_train, y_test, categorical_features, \
            continuous_features, target_name, classes, \
            feature_columns, feature_range_keys = \
            adult_data

        data_train_copy = data_train.copy()
        data_test_copy = data_test.copy()

        if missing_value_combination == \
            MISSING_VALUE.TRAIN_ONLY_MISSING_VALUES or \
            missing_value_combination == \
                MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES:
            data_train_copy.loc[data_train_copy['age'] > 30, 'age'] = np.nan

        if missing_value_combination == \
            MISSING_VALUE.TEST_ONLY_MISSING_VALUES or \
            missing_value_combination == \
                MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES:
            data_test_copy.loc[data_test_copy['age'] > 30, 'age'] = np.nan

        if categorical_missing_values:
            if missing_value_combination == \
                MISSING_VALUE.TRAIN_ONLY_MISSING_VALUES or \
                missing_value_combination == \
                    MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES:
                data_train_copy.loc[
                    data_train_copy[
                        'workclass'] == 'Private', 'workclass'] = np.nan

            if missing_value_combination == \
                MISSING_VALUE.TEST_ONLY_MISSING_VALUES or \
                missing_value_combination == \
                    MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES:
                data_test_copy.loc[
                    data_test_copy[
                        'workclass'] == 'Private', 'workclass'] = np.nan

        X_train = data_train_copy.drop([target_name], axis=1)
        X_test = data_test_copy.drop([target_name], axis=1)

        model = create_complex_classification_pipeline(
            X_train, y_train, continuous_features,
            categorical_features)

        if wrapper:
            all_data = pd.concat(
                [X_test, X_train])
            model_predict_output = model.predict(all_data)
            model_predict_proba_output = model.predict_proba(all_data)
            model_wrapper = PredictionsModelWrapperClassification(
                all_data,
                model_predict_output,
                model_predict_proba_output,
                should_construct_pandas_query=False)
            model = model_wrapper

        rai_insights = RAIInsights(
            model, data_train_copy, data_test_copy, target_name,
            task_type="classification",
            feature_metadata=FeatureMetadata(
                categorical_features=categorical_features))

        if manager_type == ManagerNames.EXPLAINER:
            if not categorical_missing_values:
                rai_insights.explainer.add()
                rai_insights.compute()
                assert len(rai_insights.explainer.get()) == 1
            else:
                if missing_value_combination != \
                        MISSING_VALUE.NO_MISSING_VALUES:
                    if missing_value_combination == \
                        MISSING_VALUE.TRAIN_ONLY_MISSING_VALUES or \
                        missing_value_combination == \
                            MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES:
                        error_message = \
                            "Categorical features workclass cannot have " + \
                            "missing values for computing explanations. " + \
                            "Please check your training data."
                    else:
                        error_message = \
                            "Categorical features workclass cannot have " + \
                            "missing values for computing explanations. " + \
                            "Please check your test data."
                    with pytest.raises(
                            UserConfigValidationException,
                            match=error_message):
                        rai_insights.explainer.add()
                else:
                    rai_insights.explainer.add()
                    rai_insights.compute()
                    assert len(rai_insights.explainer.get()) == 1
        elif manager_type == ManagerNames.ERROR_ANALYSIS:
            rai_insights.error_analysis.add()
            rai_insights.compute()
            assert len(rai_insights.error_analysis.get()) == 1
        elif manager_type == ManagerNames.COUNTERFACTUAL:
            if not wrapper:
                if missing_value_combination != \
                        MISSING_VALUE.NO_MISSING_VALUES:
                    if missing_value_combination == \
                        MISSING_VALUE.TRAIN_ONLY_MISSING_VALUES or \
                        missing_value_combination == \
                            MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES:
                        error_message = \
                            'Missing values are not allowed in ' + \
                            'the train dataset while computing ' + \
                            'counterfactuals.'
                    else:
                        error_message = \
                            'Missing values are not allowed in ' + \
                            'the test dataset while computing ' + \
                            'counterfactuals.'
                    with pytest.raises(
                            UserConfigValidationException,
                            match=error_message):
                        rai_insights.counterfactual.add(
                            total_CFs=10, desired_class="opposite")
                else:
                    rai_insights.counterfactual.add(
                        total_CFs=10, desired_class="opposite")
                    rai_insights.compute()
                    assert len(rai_insights.counterfactual.get()) == 1
        elif manager_type == ManagerNames.CAUSAL:
            if missing_value_combination != \
                    MISSING_VALUE.NO_MISSING_VALUES:
                if missing_value_combination == \
                    MISSING_VALUE.TRAIN_ONLY_MISSING_VALUES or \
                    missing_value_combination == \
                        MISSING_VALUE.BOTH_TRAIN_TEST_MISSING_VALUES:
                    error_message = \
                        'Missing values are not allowed in the ' + \
                        'train dataset while computing causal effects.'
                else:
                    error_message = \
                        'Missing values are not allowed in the ' + \
                        'test dataset while computing causal effects.'
                with pytest.raises(
                        UserConfigValidationException, match=error_message):
                    rai_insights.causal.add(treatment_features=['age'])
            else:
                rai_insights.causal.add(treatment_features=['age'])
                rai_insights.compute()
                assert len(rai_insights.causal.get()) == 1



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Manager for causal analysis."""
from pathlib import Path
from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
from econml.solutions.causal_analysis import CausalAnalysis

from raiutils.exceptions import UserConfigValidationException
from raiutils.models import ModelTask
from responsibleai._data_validations import validate_train_test_categories
from responsibleai._internal.constants import (CausalManagerKeys,
                                               ListProperties, ManagerNames)
from responsibleai._tools.causal.causal_config import CausalConfig
from responsibleai._tools.causal.causal_constants import (DefaultParams,
                                                          ModelTypes)
from responsibleai._tools.causal.causal_result import CausalResult
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.feature_metadata import FeatureMetadata
from responsibleai.managers.base_manager import BaseManager
from responsibleai.utils import (_find_features_having_missing_values,
                                 _measure_time)


class CausalManager(BaseManager):
    """Manager for generating causal analyses from a dataset."""

    def __init__(
        self,
        train: pd.DataFrame,
        test: pd.DataFrame,
        target_column: str,
        task_type: str,
        categorical_features: Optional[List[str]],
        feature_metadata: Optional[FeatureMetadata] = None
    ):
        """Construct a CausalManager for generating causal analyses
            from a dataset.
        :param train: Dataset on which to compute global causal effects
            (#samples x #features).
        :type train: pandas.DataFrame
        :param test: Dataset on which to compute local causal effects
            (#samples x #features).
        :type test: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param task_type: Task type is either 'classification/regression'
        :type task_type: str
        :param categorical_features: All categorical feature names.
        :type categorical_features: list
        :param feature_metadata: Feature metadata for the train/test
                                 dataset to identify different kinds
                                 of features in the dataset.
        :type feature_metadata: FeatureMetadata
        """
        self._train = train
        self._test = test
        self._target_column = target_column
        self._task_type = task_type

        self._categorical_features = categorical_features
        if categorical_features is None:
            self._categorical_features = []
        self._feature_metadata = feature_metadata

        self._results = []

    def add(
        self,
        treatment_features: List[str],
        heterogeneity_features: Optional[List[str]] = None,
        nuisance_model: str = ModelTypes.LINEAR,
        heterogeneity_model: str = ModelTypes.LINEAR,
        alpha: float = DefaultParams.DEFAULT_ALPHA,
        upper_bound_on_cat_expansion: int = (
            DefaultParams.DEFAULT_MAX_CAT_EXPANSION),
        treatment_cost: Union[float, List[Union[float, np.ndarray]]] = (
            DefaultParams.DEFAULT_TREATMENT_COST),
        min_tree_leaf_samples: int = (
            DefaultParams.DEFAULT_MIN_TREE_LEAF_SAMPLES),
        max_tree_depth: int = (
            DefaultParams.DEFAULT_MAX_TREE_DEPTH),
        skip_cat_limit_checks: bool = (
            DefaultParams.DEFAULT_SKIP_CAT_LIMIT_CHECKS),
        categories: Union[str, List[Union[str, List[Any]]]] = (
            DefaultParams.DEFAULT_CATEGORIES),
        n_jobs: int = (
            DefaultParams.DEFAULT_N_JOBS),
        verbose: int = (
            DefaultParams.DEFAULT_VERBOSE),
        random_state: Optional[Union[int, np.random.RandomState]] = (
            DefaultParams.DEFAULT_RANDOM_STATE),
    ):
        """Compute causal insights.
        :param treatment_features: Treatment feature names.
        :type treatment_features: list
        :param heterogeneity_features: Features that mediate the causal effect.
        :type heterogeneity_features: list
        :param nuisance_model: This model used to estimate the outcome and the
            treatment features from the other features present in user data.
            It is one of {'linear', 'automl'}, optional (default='linear').
            If 'linear', then LassoCV (for regression) and
            LogisticRegressionCV (for classification) are used.
            If 'automl', then a k-fold cross-validation and model selection
            is performed among several models and the best is chosen.
        :type nuisance_model: str
        :param heterogeneity_model: The heterogeneity model is used to
            estimate the treatment effect based on the heterogeneity features.
            It is one of {'linear', 'forest'} (default='linear').
            'linear' means that a heterogeneity model of the form
            theta(X)=<a, X> will be used, while 'forest' means that a
            forest model will be trained instead.
        :type heterogeneity_model: str
        :param alpha: Confidence level of confidence intervals.
        :type alpha: float
        :param upper_bound_on_cat_expansion: Maximum expansion for
                                             categorical features.
        :type upper_bound_on_cat_expansion: int
        :param treatment_cost: Cost of treatment. If 0, all treatments will
            have zero cost. If a list is passed, then each element will be
            applied to each treatment feature. Each element can be a scalar
            value to indicate a constant cost of applying that treatment or
            an array indicating the cost for each sample. If the treatment
            is a discrete treatment, then the array for that feature should
            be two dimensional with the first dimension representing samples
            and the second representing the difference in cost between the
            non-default values and the default value.
        :type treatment_cost: None, List of float or array
        :param min_tree_leaf_samples: Minimum number of samples per leaf
            in policy tree.
        :type min_tree_leaf_samples: int
        :param max_tree_depth: Maximum depth of policy tree.
        :type max_tree_depth: int
        :param skip_cat_limit_checks:
            By default, categorical features need to have several instances
            of each category in order for a model to be fit robustly.
            Setting this to True will skip these checks.
        :type skip_cat_limit_checks: bool
        :param categories: 'auto' or list of category values, default 'auto'
            What categories to use for the categorical columns.
            If 'auto', then the categories will be inferred for all
            categorical columns. Otherwise, this argument should have
            as many entries as there are categorical columns.
            Each entry should be either 'auto' to infer the values for
            that column or the list of values for the column.
            If explicit values are provided, the first value is treated
            as the "control" value for that column against which other
            values are compared.
        :type categories: str or list
        :param n_jobs: Degree of parallelism to use when training models
            via joblib.Parallel
        :type n_jobs: int
        :param verbose: Controls the verbosity when fitting and predicting.
        :type verbose: int
        :param random_state: Controls the randomness of the estimator.
        :type random_state: int or RandomState or None
        """
        if any(_find_features_having_missing_values(self._train)):
            raise UserConfigValidationException(
                'Missing values are not allowed in '
                'the train dataset while computing causal effects.')
        if any(_find_features_having_missing_values(self._test)):
            raise UserConfigValidationException(
                'Missing values are not allowed in '
                'the test dataset while computing causal effects.')
        if not isinstance(treatment_features, list):
            raise UserConfigValidationException(
                "Expecting a list for treatment_features but got {0}".format(
                    type(treatment_features)))
        if len(treatment_features) == 0:
            raise UserConfigValidationException(
                "Please specify at least one feature in "
                "treatment_features list")
        for feature in treatment_features:
            if self._feature_metadata and \
                    self._feature_metadata.dropped_features and \
                    feature in set(self._feature_metadata.dropped_features):
                message = ("'{}' in treatment_features has been dropped "
                           "during training the model").format(feature)
                raise UserConfigValidationException(message)
        difference_set = set(treatment_features) - set(self._train.columns)
        if len(difference_set) > 0:
            message = ("Feature names in treatment_features do "
                       f"not exist in train data: {list(difference_set)}")
            raise UserConfigValidationException(message)

        if heterogeneity_features is not None:
            difference_set = \
                set(heterogeneity_features) - set(self._train.columns)
            if len(difference_set) > 0:
                message = ("Feature names in heterogeneity_features do "
                           f"not exist in train data: {list(difference_set)}")
                raise UserConfigValidationException(message)

        if self._task_type == ModelTask.CLASSIFICATION:
            is_multiclass = len(np.unique(
                self._train[self._target_column].values).tolist()) > 2
            if is_multiclass:
                raise UserConfigValidationException(
                    "Multiclass classification isn't supported")

        if nuisance_model not in [ModelTypes.AUTOML,
                                  ModelTypes.LINEAR]:
            message = (f"nuisance_model should be one of "
                       f"['{ModelTypes.AUTOML}', "
                       f"'{ModelTypes.LINEAR}'], "
                       f"got {nuisance_model}")
            raise UserConfigValidationException(message)

        if heterogeneity_model not in [ModelTypes.FOREST,
                                       ModelTypes.LINEAR]:
            message = (f"heterogeneity_model should be one of "
                       f"['{ModelTypes.FOREST}', "
                       f"'{ModelTypes.LINEAR}'], "
                       f"got {heterogeneity_model}")
            raise UserConfigValidationException(message)

        # Check treatment_cost is valid
        if isinstance(treatment_cost, int) and \
                treatment_cost == 0:
            revised_treatment_cost = [0] * len(
                treatment_features)
        else:
            revised_treatment_cost = treatment_cost

        if not isinstance(revised_treatment_cost, list):
            message = (
                "treatment_cost must be a list with "
                "the same number of elements as "
                "treatment_features where each element "
                "is either a constant cost of treatment "
                "or an array specifying the cost of "
                "treatment per sample. "
                "Found treatment_cost of type "
                f"{type(revised_treatment_cost)}, expected list.")
            raise UserConfigValidationException(message)
        elif len(revised_treatment_cost) != \
                len(treatment_features):
            message = (
                "treatment_cost must be a list with "
                "the same number of elements as "
                "treatment_features. "
                "Length of treatment_cost was "
                f"{len(revised_treatment_cost)}, expected "
                f"{len(treatment_features)}.")
            raise UserConfigValidationException(message)

        validate_train_test_categories(
            train_data=self._train,
            test_data=self._test,
            rai_compute_type='Causal analysis',
            categoricals=self._categorical_features)

        result = CausalResult()
        result.config = CausalConfig(
            treatment_features=treatment_features,
            heterogeneity_features=heterogeneity_features,
            nuisance_model=nuisance_model,
            heterogeneity_model=heterogeneity_model,
            alpha=alpha,
            upper_bound_on_cat_expansion=upper_bound_on_cat_expansion,
            treatment_cost=revised_treatment_cost,
            min_tree_leaf_samples=min_tree_leaf_samples,
            max_tree_depth=max_tree_depth,
            skip_cat_limit_checks=skip_cat_limit_checks,
            n_jobs=n_jobs,
            categories=categories,
            verbose=verbose,
            random_state=random_state,
            categorical_features=self._categorical_features,
        )
        self._results.append(result)

    def _fit_causal_analysis(
        self,
        causal_analysis,
        X,
        y,
        max_cat_expansion
    ):
        try:
            causal_analysis.fit(X, y)
        except ValueError as e:
            message = str(e)
            expected = "increase the upper_bound_on_cat_expansion"
            clarification = (
                " Increase the value {} in rai_insights.causal.add("
                "upper_bound_on_cat_expansion={})."
            ).format(max_cat_expansion, max_cat_expansion)
            if expected in message:
                raise UserConfigValidationException(
                    message + clarification)
            raise e

    def _create_policy(
        self,
        causal_result,
        X_test,
        treatment_feature,
        treatment_cost,
        alpha,
        max_tree_depth,
        min_tree_leaf_samples,
    ):
        return causal_result._create_policy(
            X_test=X_test,
            treatment_feature=treatment_feature,
            treatment_cost=treatment_cost,
            alpha=alpha,
            max_tree_depth=max_tree_depth,
            min_tree_leaf_samples=min_tree_leaf_samples,
        )

    def _whatif(self, id, X, X_feature_new, feature_name, y, alpha=0.1):
        """Get what-if data."""
        filtered = [r for r in self.get() if r.id == id]
        if len(filtered) == 0:
            raise ValueError(f"Failed to find causal result with ID: {id}")
        result = filtered[0]
        return result._whatif(X, X_feature_new, feature_name,
                              y, alpha=alpha).to_dict(orient="records")

    def request_global_cohort_effects(self, id, X_test):
        """Get global causal effects for cohort data.

        :param id: The query id for finding the
                   causal config.
        :type id: str
        :param X_test: The data for which the global causal effects
                       needs to be generated.
        :type X_test: Any
        :return: An object of type CausalData with
                 causal effects.
        :rtype: CausalData

        """
        filtered = [r for r in self.get() if r.id == id]
        if len(filtered) == 0:
            raise ValueError(f"Failed to find causal result with ID: {id}")
        result = filtered[0]
        return result._global_cohort_effects(X_test)

    def request_local_instance_effects(self, id, X_test):
        """Get local causal effects for a given data point.

        :param id: The query id for finding the
                   causal config.
        :type id: str
        :param X_test: The data for which the local causal effects
                       needs to be generated for a given point.
        :type X_test: Any
        :return: An object of type CausalData with
                 causal effects for a given point.
        :rtype: CausalData
        """
        filtered = [r for r in self.get() if r.id == id]

        if len(filtered) == 0:
            raise ValueError(f"Failed to find causal result with ID: {id}")

        if not isinstance(X_test, pd.DataFrame):
            raise UserConfigValidationException(
                'Data is of type {0} but it must be '
                'a pandas DataFrame.'.format(type(X_test)))

        if X_test.shape[0] > 1:
            raise UserConfigValidationException(
                'Only one row of data is allowed for '
                'local causal effects.')

        result = filtered[0]
        return result._local_instance_effects(X_test)

    def request_global_cohort_policy(self, id, X_test):
        """Get global causal policy for cohort data.

        :param id: The query id for finding the
                   causal config.
        :type id: str
        :param X_test: The data for which the causal policy
                       needs to be generated.
        :type X_test: Any
        :return: An object of type CausalData with
                 causal effects.
        :rtype: CausalData
        """
        filtered = [r for r in self.get() if r.id == id]
        if len(filtered) == 0:
            raise ValueError(f"Failed to find causal result with ID: {id}")
        result = filtered[0]
        return result._global_cohort_policy(X_test)

    @_measure_time
    def compute(self):
        """Computes the causal effects by running the causal
           configuration."""
        print("Causal Effects")
        print('Current Status: Generating Causal Effects.')
        is_classification = self._task_type == ModelTask.CLASSIFICATION
        for result in self._results:
            causal_config = result.config
            if not result.is_computed:
                analysis = CausalAnalysis(
                    causal_config.treatment_features,
                    self._categorical_features,
                    heterogeneity_inds=causal_config.heterogeneity_features,
                    classification=is_classification,
                    nuisance_models=causal_config.nuisance_model,
                    heterogeneity_model=causal_config.heterogeneity_model,
                    upper_bound_on_cat_expansion=causal_config.
                    upper_bound_on_cat_expansion,
                    skip_cat_limit_checks=causal_config.skip_cat_limit_checks,
                    n_jobs=causal_config.n_jobs,
                    categories=causal_config.categories,
                    verbose=causal_config.verbose,
                    random_state=causal_config.random_state,
                )

                X_train = self._train.drop([self._target_column], axis=1)
                X_test = self._test.drop([self._target_column], axis=1)
                y_train = self._train[self._target_column].values.ravel()

                self._fit_causal_analysis(
                    analysis, X_train, y_train,
                    causal_config.upper_bound_on_cat_expansion)
                result.causal_analysis = analysis

                result.global_effects = analysis.global_causal_effect(
                    alpha=causal_config.alpha, keep_all_levels=True)
                result.local_effects = analysis.local_causal_effect(
                    X_test, alpha=causal_config.alpha, keep_all_levels=True)

                result.policies = []
                for i in range(len(causal_config.treatment_features)):
                    policy = self._create_policy(
                        result, X_test,
                        causal_config.treatment_features[i],
                        causal_config.treatment_cost[i],
                        causal_config.alpha, causal_config.max_tree_depth,
                        causal_config.min_tree_leaf_samples)
                    result.policies.append(policy)

                result._validate_schema()
        print('Current Status: Finished generating causal effects.')

    def get(self):
        """Get the computed causal insights."""
        return self._results

    def list(self):
        """List information about the CausalManager.

        :return: A dictionary of properties.
        :rtype: dict
        """
        props = {ListProperties.MANAGER_TYPE: self.name}
        causal_props_list = []
        for result in self._results:
            causal_config_dict = result.config.get_config_as_dict()
            causal_config_dict[CausalManagerKeys.GLOBAL_EFFECTS_COMPUTED] = \
                True if result.global_effects is not None else False
            causal_config_dict[CausalManagerKeys.LOCAL_EFFECTS_COMPUTED] = \
                True if result.local_effects is not None else False
            causal_config_dict[CausalManagerKeys.POLICIES_COMPUTED] = \
                True if result.policies is not None else False
            causal_props_list.append(causal_config_dict)
        props[CausalManagerKeys.CAUSAL_EFFECTS] = causal_props_list

        return props

    def get_data(self):
        """Get causal data
        :return: List of CausalData objects.
        :rtype: List[CausalData]
        """
        return [result._get_dashboard_object() for result in self._results]

    @property
    def name(self):
        """Get the name of the causal manager.
        :return: The name of the causal manager.
        :rtype: str
        """
        return ManagerNames.CAUSAL

    def _save(self, path):
        """Save the CausalManager to the given path.

        :param path: The directory path to save the CausalManager to.
        :type path: str
        """
        causal_dir = Path(path)
        causal_dir.mkdir(parents=True, exist_ok=True)

        # Save results to disk
        for result in self._results:
            directory_manager = DirectoryManager(parent_directory_path=path)
            data_path = directory_manager.create_data_directory()
            result.save(data_path)

    @staticmethod
    def _load(path, rai_insights):
        """Load the CausalManager from the given path.

        :param path: The directory path to load the CausalManager from.
        :type path: str
        :param rai_insights: The loaded parent RAIInsights.
        :type rai_insights: RAIInsights
        :return: The CausalManager manager after loading.
        :rtype: CausalManager
        """
        inst = CausalManager.__new__(CausalManager)

        # Rehydrate results
        all_causal_dirs = DirectoryManager.list_sub_directories(path)
        inst.__dict__['_results'] = []
        for causal_dir in all_causal_dirs:
            dm = DirectoryManager(parent_directory_path=path,
                                  sub_directory_name=causal_dir)
            causal_result = CausalResult.load(dm.get_data_directory())
            inst.__dict__['_results'].append(causal_result)

        # Rehydrate model analysis data
        inst.__dict__['_train'] = rai_insights.train
        inst.__dict__['_test'] = rai_insights.test
        inst.__dict__['_target_column'] = rai_insights.target_column
        inst.__dict__['_task_type'] = rai_insights.task_type
        inst.__dict__['_categorical_features'] = \
            rai_insights.categorical_features
        inst.__dict__['_feature_metadata'] = rai_insights._feature_metadata

        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import itertools
import warnings
from typing import Callable, Dict, List, Tuple

import numpy as np
import pandas as pd
import scipy

from responsibleai.databalanceanalysis.balance_measures import BalanceMeasures
from responsibleai.databalanceanalysis.constants import Constants, Measures


def get_statistical_parity(
    p_pos: float, p_feature: float, p_pos_feature: float, total_count: int
) -> float:
    """
    Returns the Statistical Parity of a feature, a fairness measure which
    states the probability that the feature sees the positive outcome.
    https://en.wikipedia.org/wiki/Fairness_%28machine_learning%29

    :param p_pos: The probability of the positive outcome in the dataset.
    :type p_pos: float
    :param p_feature: The probability of the feature in the dataset.
    :type p_feature: float
    :param p_pos_feature: The probability of the positive outcome and the
        feature in the dataset.
    :type p_pos_feature: float
    :param total_count: The total number of observations in the dataset.
    :type total_count: int
    :return: The Statistical Parity of the feature.
    :rtype: float
    """
    return p_pos_feature / p_feature


def get_point_mutual(
    p_pos: float, p_feature: float, p_pos_feature: float, total_count: int
) -> float:
    """
    Returns the Point Mutual Information of a feature, an entropy measure
    which quantifies the discrepancy between the probability of their
    coincidence given their joint distribution and their individual
    distributions (assuming independence).
    https://en.wikipedia.org/wiki/Pointwise_mutual_information

    :param p_pos: The probability of the positive outcome in the dataset.
    :type p_pos: float
    :param p_feature: The probability of the feature in the dataset.
    :type p_feature: float
    :param p_pos_feature: The probability of the positive outcome and the
        feature in the dataset.
    :type p_pos_feature: float
    :param total_count: The total number of observations in the dataset.
    :type total_count: int
    :return: The Point Mutual Information of the feature.
    :rtype: float
    """
    dp = get_statistical_parity(p_pos, p_feature, p_pos_feature, total_count)
    return -np.inf if dp == 0 else np.log(dp)


def get_sorenson_dice(
    p_pos: float, p_feature: float, p_pos_feature: float, total_count: int
) -> float:
    """
    Returns the Sørensen-Dice coefficient of a feature, an
    intersection-over-union measure used to gauge the similarity between
    two samples. Also known as the F1 score.
    https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient

    :param p_pos: The probability of the positive outcome in the dataset.
    :type p_pos: float
    :param p_feature: The probability of the feature in the dataset.
    :type p_feature: float
    :param p_pos_feature: The probability of the positive outcome and the
        feature in the dataset.
    :type p_pos_feature: float
    :param total_count: The total number of observations in the dataset.
    :type total_count: int
    :return: The Sørensen-Dice coefficient of the feature.
    :rtype: float
    """
    return p_pos_feature / (p_feature + p_pos)


def get_jaccard_index(
    p_pos: float, p_feature: float, p_pos_feature: float, total_count: int
) -> float:
    """
    Returns the Jaccard Index of a feature, an
    intersection-over-union measure used for gauging similarity and
    diversity of sample sets.
    https://en.wikipedia.org/wiki/Jaccard_index

    :param p_pos: The probability of the positive outcome in the dataset.
    :type p_pos: float
    :param p_feature: The probability of the feature in the dataset.
    :type p_feature: float
    :param p_pos_feature: The probability of the positive outcome and the
        feature in the dataset.
    :type p_pos_feature: float
    :param total_count: The total number of observations in the dataset.
    :type total_count: int
    :return: The Jaccard Index of the feature.
    :rtype: float
    """
    return p_pos_feature / (p_feature + p_pos - p_pos_feature)


def get_kr_correlation(
    p_pos: float, p_feature: float, p_pos_feature: float, total_count: int
) -> float:
    """
    Returns the Kendall Rank Correlation of a feature, a correlation
    and statistical test that measures the ordinal association between two
    measured quantities.
    https://en.wikipedia.org/wiki/Kendall_rank_correlation_coefficient

    :param p_pos: The probability of the positive outcome in the dataset.
    :type p_pos: float
    :param p_feature: The probability of the feature in the dataset.
    :type p_feature: float
    :param p_pos_feature: The probability of the positive outcome and the
        feature in the dataset.
    :type p_pos_feature: float
    :param total_count: The total number of observations in the dataset.
    :type total_count: int
    :return: The Kendall Rank Correlation of the feature.
    :rtype: float
    """
    a = np.power(total_count, 2) * (
        1 -
        2 * p_feature -
        2 * p_pos +
        2 * p_pos_feature +
        2 * p_pos * p_feature
    )
    b = total_count * (2 * p_feature + 2 * p_pos - 4 * p_pos_feature - 1)
    c = np.power(total_count, 2) * np.sqrt(
        (p_feature - np.power(p_feature, 2)) * (p_pos - np.power(p_pos, 2))
    )
    return (a + b) / c


def get_log_likelihood_ratio(
    p_pos: float, p_feature: float, p_pos_feature: float, total_count: int
) -> float:
    """
    Returns the Log likelihood ratio of a feature, a correlation
    and statistical test that gives the probability of correctly predicting
    the label in ratio to probability of incorrectly predicting the label.
    https://en.wikipedia.org/wiki/Likelihood_function#Likelihood_ratio

    :param p_pos: The probability of the positive outcome in the dataset.
    :type p_pos: float
    :param p_feature: The probability of the feature in the dataset.
    :type p_feature: float
    :param p_pos_feature: The probability of the positive outcome and the
        feature in the dataset.
    :type p_pos_feature: float
    :param total_count: The total number of observations in the dataset.
    :type total_count: int
    :return: The Log likelihood ratio of the feature.
    :rtype: float
    """
    return -np.inf if p_pos_feature == 0 else np.log(p_pos_feature / p_pos)


def get_ttest_stat(
    p_pos: float, p_feature: float, p_pos_feature: float, total_count: int
) -> float:
    """
    Returns the t-test statistic of a feature, a correlation
    and statistical test that is used to compare the means of two groups
    (pairwise). t-test compares the means of two groups and
    this gets a value to be looked up on t-distribution.
    https://en.wikipedia.org/wiki/Student%27s_t-test

    :param p_pos: The probability of the positive outcome in the dataset.
    :type p_pos: float
    :param p_feature: The probability of the feature in the dataset.
    :type p_feature: float
    :param p_pos_feature: The probability of the positive outcome and the
        feature in the dataset.
    :type p_pos_feature: float
    :param total_count: The total number of observations in the dataset.
    :type total_count: int
    :return: The t-test statistic of the feature.
    :rtype: float
    """
    return (p_pos - (p_feature * p_pos)) / np.sqrt(p_feature * p_pos)


def get_ttest_pvalue(metric: float, num_unique_vals: int) -> float:
    """
    Returns the t-test p-value. Once the t-test statistic and degrees
    of freedom are determined, the p-value can be found using a table
    of values from the t-distribution.
    https://en.wikipedia.org/wiki/Student%27s_t-test

    :param metric: The t-test statistic of the feature.
    :type metric: float
    :param num_unique_vals: The number of unique values in the feature.
    :type num_unique_vals: int
    :return: The t-test p-value of the feature.
    :rtype: float
    """
    t_statistic = metric
    dof = num_unique_vals - 1
    return scipy.stats.t.sf(abs(t_statistic), dof)


class FeatureBalanceMeasures(BalanceMeasures):
    """
    This class computes a set of feature balance measures that allow us to see
    whether each combination of a feature is receiving the positive outcome
    (true prediction) at balanced probability.
    """

    CLASS_A = "ClassA"
    CLASS_B = "ClassB"
    PROB_POSITIVE = "p_pos"
    PROB_FEATURE = "p_feature"
    PROB_POS_FEATURE = "p_pos_feature"

    FEATURE_METRICS: Dict[
        Measures, Callable[[float, float, float, int], float]
    ] = {
        Measures.STATISTICAL_PARITY: get_statistical_parity,
        Measures.POINTWISE_MUTUAL_INFO: get_point_mutual,
        Measures.SD_COEF: get_sorenson_dice,
        Measures.JACCARD_INDEX: get_jaccard_index,
        Measures.KR_CORRELATION: get_kr_correlation,
        Measures.LOG_LIKELIHOOD: get_log_likelihood_ratio,
        Measures.TTEST: get_ttest_stat,
    }

    OVERALL_METRICS: Dict[
        Tuple[Measures, Measures], Callable[[float, int], float]
    ] = {
        (
            Measures.TTEST_PVALUE,
            Measures.TTEST,
        ): get_ttest_pvalue,
    }

    def __init__(
        self, cols_of_interest: List[str], label_col: str, pos_label: str
    ) -> None:
        """
        Construct a FeatureBalanceMeasures class for calculating
        feature balance measures on a dataset.

        :param cols_of_interest: The list of columns to compute
            feature balance measures on.
        :type cols_of_interest: List[str]
        :param label_col: The name of the label column.
        :type label_col: str
        :param pos_label: The label value that denotes a positive label.
        :type pos_label: str
        """
        super().__init__(cols_of_interest=cols_of_interest)
        self.label_col = label_col
        self.pos_label = pos_label

    def measures(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """
        Returns feature balance measures for the given dataset,
        columns of interest, and label column.

        The following measures are computed:
        * Statistical Parity -
            https://en.wikipedia.org/wiki/Fairness_%28machine_learning%29
        * Pointwise Mutual Information -
            https://en.wikipedia.org/wiki/Pointwise_mutual_information
        * Sorensen-Dice Coefficient -
            https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
        * Jaccard Index -
            https://en.wikipedia.org/wiki/Jaccard_index
        * Kendall Rank Correlation -
            https://en.wikipedia.org/wiki/Kendall_rank_correlation_coefficient
        * Log-Likelihood Ratio -
            https://en.wikipedia.org/wiki/Likelihood_function#Likelihood_ratio
        * t-test -
            https://en.wikipedia.org/wiki/Student%27s_t-test

        :param dataset: The dataset to compute the measures on.
        :type df: pd.DataFrame
        :return: A pandas DataFrame that contains four columns and has the
            format: <Feature Name>, <Class A value>, <Class B value>, <dict of
            measure name -> measure value key-value pairs>
        :rtype: pd.DataFrame
        """
        feature_balance_measures = [
            FeatureBalanceMeasures._get_measure_gaps_for_col(
                df=dataset,
                col_of_interest=col,
                label_col=self.label_col,
                pos_label=self.pos_label,
            )
            for col in self.cols_of_interest
        ]
        return pd.concat(feature_balance_measures)

    @staticmethod
    def _get_measure_gaps_for_col(
        df: pd.DataFrame, col_of_interest: str, label_col: str, pos_label: str
    ) -> pd.DataFrame:
        """
        For the column of interest, computes "gaps" between two classes
        (feature values) for every combination of values in the column.
        Returns a dataframe that contains as many rows as there are unique
        combinations in the column of interest and for each row, the
        two feature values being compared and a dictionary of measure values.

        :param df: The dataset to compute gaps between feature values on.
        :type df: pd.DataFrame
        :param col_of_interest: The column of interest to compute gaps between
            feature values on.
        :type col_of_interest: str
        :param label_col: The name of the label column.
        :type label_col: str
        :param pos_label: The label value that denotes a positive label.
        :type pos_label: str
        :return: A dataframe that contains four columns.
        :rtype: pd.DataFrame
        """
        unique_vals = df[col_of_interest].unique()

        if unique_vals.size < 2:
            warnings.warn((f"Column '{col_of_interest}' has less than 2"
                           " unique values, so feature balance measures for"
                           f" {label_col}=={pos_label} are not being computed"
                           " for this column."))
            return pd.DataFrame()

        # create list of tuples of the pairings of classes
        pairs = list(itertools.combinations(unique_vals, 2))
        gap_df = pd.DataFrame(
            pairs,
            columns=[
                FeatureBalanceMeasures.CLASS_A,
                FeatureBalanceMeasures.CLASS_B,
            ],
        )
        gap_df[Constants.FEATURE_NAME.value] = col_of_interest

        metrics_df = FeatureBalanceMeasures._get_individual_measures(
            df=df,
            col_of_interest=col_of_interest,
            label_col=label_col,
            pos_label=pos_label,
        )

        for measure in FeatureBalanceMeasures.FEATURE_METRICS.keys():
            classA_metric = gap_df[FeatureBalanceMeasures.CLASS_A].apply(
                lambda row: metrics_df.loc[row]
            )[measure.value]
            classB_metric = gap_df[FeatureBalanceMeasures.CLASS_B].apply(
                lambda row: metrics_df.loc[row]
            )[measure.value]
            gap_df[measure.value] = classA_metric - classB_metric

        # For overall stats
        for (
            measure,
            test_stat,
        ), func in FeatureBalanceMeasures.OVERALL_METRICS.items():
            gap_df[measure.value] = gap_df[test_stat.value].apply(
                lambda row: func(row, len(unique_vals))
            )

        return gap_df

    @staticmethod
    def _get_individual_measures(
        df: pd.DataFrame, col_of_interest: str, label_col: str, pos_label: str
    ):
        """
        For the column of interest, computes the individual feature balance
        measures for every unique value in the column. These measures are
        used to compute "gaps" between two classes (feature values).

        :param df: The df to compute individual feature balance
            measures on.
        :type df: pd.DataFrame
        :param col_of_interest: The column of interest to compute the
            individual feature balance measures on.
        :type col_of_interest: str
        :param label_col: The name of the label column.
        :type label_col: str
        :param pos_label: The label value that denotes a positive label.
        :type pos_label: str
        :return: A dataframe that contains individual feature balance measures.
        :rtype: pd.DataFrame
        """
        num_rows = df.shape[0]

        p_feature_col = (
            df[col_of_interest]
            .value_counts()
            .rename(FeatureBalanceMeasures.PROB_FEATURE) / num_rows
        )
        p_pos_feature_col = (
            df[
                df[label_col].astype("str") == str(pos_label)
            ][col_of_interest]
            .value_counts()
            .rename(FeatureBalanceMeasures.PROB_POS_FEATURE) / num_rows
        ).fillna(0)

        new_df = pd.concat([p_feature_col, p_pos_feature_col], axis=1)
        new_df[FeatureBalanceMeasures.PROB_POSITIVE] = (
            df[
                df[label_col].astype("str") == str(pos_label)
            ].shape[0] / num_rows
        )

        for measure, func in FeatureBalanceMeasures.FEATURE_METRICS.items():
            new_df[measure.value] = new_df.apply(
                lambda row: func(
                    row[FeatureBalanceMeasures.PROB_POSITIVE],
                    row[FeatureBalanceMeasures.PROB_FEATURE],
                    row[FeatureBalanceMeasures.PROB_POS_FEATURE],
                    num_rows,
                ),
                axis=1,
            )

        return new_df



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Tools to help understand imbalance and potential bias in data."""

from .aggregate_balance_measures import AggregateBalanceMeasures
from .distribution_balance_measures import DistributionBalanceMeasures
from .feature_balance_measures import FeatureBalanceMeasures

__all__ = [
    "AggregateBalanceMeasures",
    "FeatureBalanceMeasures",
    "DistributionBalanceMeasures",
]



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd
import pytest

from raiutils.exceptions import SystemErrorException
from responsibleai.rai_insights import RAIInsights


class TestRAIInsightsGetFeatureRanges:
    def test_get_feature_range(self):
        data = {'Category': ['A', 'B', 'C', None, 'D'],
                'Numerical': [10, 20, None, 40, 50]}

        df = pd.DataFrame(data)

        feature_ranges = RAIInsights._get_feature_ranges(
            df, ['Category'], ['Category', 'Numerical'])
        assert len(feature_ranges) == 2
        assert 'Category' == feature_ranges[0]['column_name']
        assert 'Numerical' == feature_ranges[1]['column_name']

    def test_invalid_float_cast_raises_value_error(self):
        data = {'col1': ['A', 'B', 'C', '50']}

        df = pd.DataFrame(data)

        with pytest.raises(SystemErrorException,
                           match='Unable to convert min or max value'):
            RAIInsights._get_feature_ranges(df, [], ['col1'])
